using System;
using System.Text;
using org.bouncycastle.math;
using org.bouncycastle.math.ec;
using org.bouncycastle.security;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.signers;
using org.bouncycastle.crypto.digests;
using org.bouncycastle.crypto.engines;
using org.bouncycastle.crypto.agreement;
using org.bouncycastle.crypto.generators;
using org.bouncycastle.crypto.parameters;
using org.bouncycastle.crypto.macs;
using org.bouncycastle.crypto.modes;
using org.bouncycastle.crypto.paddings;
using org.bouncycastle.util.encoders;
using org.bouncycastle.util.test;

namespace test.crypto
{
	/**
	 * test for ECIES - Elliptic Curve Integrated Encryption Scheme
	 */
	public class ECIESTest: Test
	{
		ECIESTest() {}

		public String getName()
		{
			return "ECIES";
		}

		private bool sameAs(
			byte[]  a,
			byte[]  b)
		{
			if (a.Length != b.Length)
			{
				return false;
			}

			for (int i = 0; i != a.Length; i++)
			{
				if (a[i] != b[i])
				{
					return false;
				}
			}

			return true;
		}

		public TestResult perform()
		{
			SecureRandom    random = new SecureRandom();
			ECCurve.Fp curve = new ECCurve.Fp(
				new BigInteger("883423532389192164791648750360308885314476597252960362792450860609699839"), // q
				new BigInteger("7fffffffffffffffffffffff7fffffffffff8000000000007ffffffffffc", 16), // a
				new BigInteger("6b016c3bdcf18941d0d654921475ca71a9db2fb27d1d37796185c2942c0a", 16)); // b

			ECDomainParameters parameters = new ECDomainParameters(
								   curve,
								   curve.decodePoint(Hex.decode("020ffa963cdca8816ccc33b8642bedf905c3d358573d3f27fbbd3b3cb9aaaf")), // G
								   new BigInteger("883423532389192164791648750360308884807550341691627752275345424702807307")); // n


			ECKeyPairGenerator          pGen = new ECKeyPairGenerator();
			ECKeyGenerationParameters   genParam = new ECKeyGenerationParameters(
				parameters,
				random);

			pGen.init(genParam);

			AsymmetricCipherKeyPair  p1 = pGen.generateKeyPair();
			AsymmetricCipherKeyPair  p2 = pGen.generateKeyPair();
    
			//
			// stream test
			//
			IESEngine      i1 = new IESEngine(
				new ECDHBasicAgreement(),
				new KDF2BytesGenerator(new SHA1Digest()),
				new HMac(new SHA1Digest()));
			IESEngine      i2 = new IESEngine(
				new ECDHBasicAgreement(),
				new KDF2BytesGenerator(new SHA1Digest()),
				new HMac(new SHA1Digest()));
			byte[]         d = new byte[] { 1, 2, 3, 4, 5, 6, 7, 8 };
			byte[]         e = new byte[] { 8, 7, 6, 5, 4, 3, 2, 1 };
			IESParameters  p = new IESParameters(d, e, 64);

			i1.init(true, p1.getPrivate(), p2.getPublic(), p);
			i2.init(false, p2.getPrivate(), p1.getPublic(), p);

			byte[] message = Hex.decode("1234567890abcdef");

			try
			{
				byte[]   out1 = i1.processBlock(message, 0, message.Length);

				byte[]   out2 = i2.processBlock(out1, 0, out1.Length);

				if (!sameAs(out2, message))
				{
					return new SimpleTestResult(false, this.getName() + ": stream cipher test failed");
				}
      
			}
			catch (Exception ex)
			{
				return new SimpleTestResult(false, this.getName() + ": stream cipher test exception " + ex.ToString());
			}

			//
			// twofish with IV0 test
			//
			BufferedBlockCipher c1 = new PaddedBufferedBlockCipher(
				new CBCBlockCipher(new TwofishEngine()));
			BufferedBlockCipher c2 = new PaddedBufferedBlockCipher(
				new CBCBlockCipher(new TwofishEngine()));
			i1 = new IESEngine(
				new ECDHBasicAgreement(),
				new KDF2BytesGenerator(new SHA1Digest()),
				new HMac(new SHA1Digest()),
				c1);
			i2 = new IESEngine(
				new ECDHBasicAgreement(),
				new KDF2BytesGenerator(new SHA1Digest()),
				new HMac(new SHA1Digest()),
				c2);
			d = new byte[] { 1, 2, 3, 4, 5, 6, 7, 8 };
			e = new byte[] { 8, 7, 6, 5, 4, 3, 2, 1 };
			p = new IESWithCipherParameters(d, e, 64, 128);

			i1.init(true, p1.getPrivate(), p2.getPublic(), p);
			i2.init(false, p2.getPrivate(), p1.getPublic(), p);

			message = Hex.decode("1234567890abcdef");

			try
			{
				byte[]    out1 = i1.processBlock(message, 0, message.Length);

				byte[]    out2 = i2.processBlock(out1, 0, out1.Length);

				if (!sameAs(out2, message))
				{
					return new SimpleTestResult(false, this.getName() + ": twofish cipher test failed");
				}
			}
			catch (Exception ex)
			{
				return new SimpleTestResult(false, this.getName() + ": twofish cipher test exception " + ex.ToString());
			}

			return new SimpleTestResult(true, this.getName() + ": Okay");
		}

		[STAThread]
		static void Main(string[] args)
		{
			ECIESTest    test = new ECIESTest();
			TestResult      result = test.perform();

			Console.WriteLine(result.ToString());
		}
	}
}
