using System;
using System.IO;
using System.Collections;
//using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	abstract public class ASN1Set : DERObject
	{
		protected ArrayList m_set = new ArrayList();

		/**
		* return an ASN1Set from the given object.
		*
		* @param obj the object we want converted.
		* @exception ArgumentException if the object cannot be converted.
		*/
		public static ASN1Set getInstance(
			Object  obj)
		{
			if ((obj == null) || (typeof(ASN1Set).IsInstanceOfType(obj)))
			{
				return (ASN1Set)obj;
			}

			throw new ArgumentException("unknown object in getInstance");
		}

		/**
		* Return an ASN1 set from a tagged object. There is a special
		* case here, if an object appears to have been explicitly tagged on 
		* reading but we were expecting it to be implictly tagged in the 
		* normal course of events it indicates that we lost the surrounding
		* set - so we need to add it back (this will happen if the tagged
		* object is a sequence that contains other sequences). If you are
		* dealing with implicitly tagged sets you really <b>should</b>
		* be using this method.
		*
		* @param obj the tagged object.
		* @param flagExplicit true if the object is meant to be explicitly tagged
		*          false otherwise.
		* @exception ArgumentException if the tagged object cannot
		*          be converted.
		*/
		public static ASN1Set getInstance(
			ASN1TaggedObject    obj,
			bool             flagExplicit)
		{
			if (flagExplicit)
			{
				if (!obj.isExplicit())
				{
					throw new ArgumentException("object implicit - explicit expected.");
				}

				return (ASN1Set)obj.getObject();
			}
			else
			{
				//
				// constructed object which appears to be explicitly tagged
				// and it's really implicit means we have to add the
				// surrounding sequence.
				//
				if (obj.isExplicit())
				{
					ASN1Set    aSet = new DERSet(obj.getObject());

					return aSet;
				}
				else
				{
					if (typeof(ASN1Set).IsInstanceOfType(obj.getObject()))
					{
						return (ASN1Set)obj.getObject();
					}

					//
					// in this case the parser returns a sequence, convert it
					// into a set.
					//
					ASN1EncodableVector  v = new ASN1EncodableVector();

					if (typeof(ASN1Sequence).IsInstanceOfType(obj.getObject()))
					{
						ASN1Sequence s = (ASN1Sequence)obj.getObject();
						IEnumerator e = s.getObjects();

						while (e.MoveNext())
						{
							v.add((DEREncodable)e.Current);
						}

						return new DERSet(v);
					}
				}
			}

			throw new ArgumentException(
						"unknown object in getInstanceFromTagged");
		}

		public ASN1Set()
		{
		}

		public IEnumerator getObjects()
		{
			return m_set.GetEnumerator();
		}

		/**
		* return the object at the set postion indicated by index.
		*
		* @param the set number (starting at zero) of the object
		* @return the object at the set postion indicated by index.
		*/
		public DEREncodable getObjectAt(
			int index)
		{
			return (DEREncodable)m_set[index];
		}

		/**
		* return the number of objects in this set.
		*
		* @return the number of objects in this set.
		*/
		public int size()
		{
			return m_set.Count;
		}

		public override int GetHashCode()
		{
			IEnumerator             e = this.getObjects();
			int                     hashCode = 0;

			while (e.MoveNext())
			{
				hashCode ^= e.Current.GetHashCode();
			}

			return hashCode;
		}

		public override bool Equals(
			Object  o)
		{
			if ((o == null) || !(typeof(ASN1Set).IsInstanceOfType(o)))
			{
				return false;
			}

			ASN1Set   other = (ASN1Set)o;

			if (this.size() != other.size())
			{
				return false;
			}

			IEnumerator s1 = this.getObjects();
			IEnumerator s2 = other.getObjects();

			while (s1.MoveNext())
			{
				if (!s1.Current.Equals(s2.Current))
				{
					return false;
				}
			}

			return true;
		}

		public virtual void addObject(
			DEREncodable obj)
		{
			m_set.Add(obj);
		}
	}

}