using System;
using System.IO;
//using System.Collections;
using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	public class DERBitString : DERObject
	{
		protected byte[]      data;
		protected int         padBits;

		/**
		* return the correct number of pad bits for a bit string defined in
		* a 16 bit constant
		*/
		static protected int getPadBits(
			int bitString)
		{
			int val;

			if (bitString == 0)
			{
				return 7;
			}

			if (bitString > 255)
			{
				val = ((bitString >> 8) & 0xFF);
			}
			else
			{
				val = (bitString & 0xFF);
			}

			int bits = 1;

			while (((val <<= 1) & 0xFF) != 0)
			{
				bits++;
			}

			return 8 - bits;
		}

		/**
		* return the correct number of bytes for a bit string defined in
		* a 16 bit constant
		*/
		static protected byte[] getBytes(
			int bitString)
		{
			if (bitString > 255)
			{
				byte[]  bytes = new byte[2];

				bytes[0] = (byte)(bitString & 0xFF);
				bytes[1] = (byte)((bitString >> 8) & 0xFF);

				return bytes;
			}
			else
			{
				byte[]  bytes = new byte[1];

				bytes[0] = (byte)(bitString & 0xFF);

				return bytes;
			}
		}

		/**
		* return a Bit String from the passed in object
		*
		* @exception ArgumentException if the object cannot be converted.
		*/
		public static DERBitString getInstance(
			Object  obj)
		{
			if ((obj == null) || (typeof(DERBitString).IsInstanceOfType(obj)))
			{
				return (DERBitString)obj;
			}
			if (typeof(ASN1OctetString).IsInstanceOfType(obj))
			{
				byte[]  bytes = ((ASN1OctetString)obj).getOctets();
				int     padBits = bytes[0];
				byte[]  data = new byte[bytes.Length - 1];

				Array.Copy(bytes, 1, data, 0, bytes.Length - 1);

				return new DERBitString(data, padBits);
			}

			if (typeof(ASN1TaggedObject).IsInstanceOfType(obj))
			{
				return getInstance(((ASN1TaggedObject)obj).getObject());
			}

			throw new ArgumentException("illegal object in getInstance: " + obj.GetType().ToString());
		}

		/**
		* return a Bit String from a tagged object.
		*
		* @param obj the tagged object holding the object we want
		* @param flagExplicit true if the object is meant to be explicitly
		*              tagged false otherwise.
		* @exception ArgumentException if the tagged object cannot
		*               be converted.
		*/
		public static DERBitString getInstance(
			ASN1TaggedObject obj,
			bool          flagExplicit)
		{
			return getInstance(obj.getObject());
		}
	    
		protected DERBitString(
			byte    data,
			int     padBits)
		{
			this.data = new byte[1];
			this.data[0] = data;
			this.padBits = padBits;
		}

		/**
		* @param data the octets making up the bit string.
		* @param padBits the number of extra bits at the end of the string.
		*/
		public DERBitString(
			byte[]  data,
			int     padBits)
		{
			this.data = data;
			this.padBits = padBits;
		}

		public DERBitString(
			byte[]  data) : this(data, 0){}

		public DERBitString(
			DEREncodable  obj)
		{
			try
			{
				MemoryStream   bOut = new MemoryStream();
				DEROutputStream         dOut = new DEROutputStream(bOut);

				dOut.writeObject(obj);
				dOut.close();

				this.data = bOut.ToArray();
				this.padBits = 0;
			}
			catch (IOException e)
			{
				throw new ArgumentException("Error processing object : " + e.ToString());
			}
		}

		public byte[] getBytes()
		{
			return data;
		}

		public int getPadBits()
		{
			return padBits;
		}

		public override void encode(DEROutputStream  outStream)
			//throws IOException
		{
			byte[]  bytes = new byte[getBytes().Length + 1];

			bytes[0] = (byte)getPadBits();
			Array.Copy(getBytes(), 0, bytes, 1, bytes.Length - 1);

			outStream.writeEncoded(DERTags.BIT_STRING, bytes);
		}

	    public override bool Equals(Object  o)
		{
			if ((o == null) || !(typeof(DERBitString).IsInstanceOfType(o)))
			{
				return false;
			}

			DERBitString  other = (DERBitString)o;

			if (data.Length != other.data.Length)
			{
				return false;
			}

			for (int i = 0; i != data.Length; i++)
			{
				if (data[i] != other.data[i])
				{
					return false;
				}
			}

			return (padBits == other.padBits);
		}

		public override int GetHashCode()
		{
			return base.GetHashCode();
		}

	}

}