using System;
using System.IO;
using System.Text;
using org.bouncycastle.io;

namespace org.bouncycastle.asn1
{
	public class DERObjectIdentifier: DERObject
	{
		String      identifier;

		/**
		* return an OID from the passed in object
		*
		* @exception ArgumentException if the object cannot be converted.
		*/
		public static DERObjectIdentifier getInstance(
			Object  obj)
		{
			if ((obj == null) || (typeof(DERObjectIdentifier).IsInstanceOfType(obj)))
			{
				return (DERObjectIdentifier)obj;
			}

			if (typeof(ASN1OctetString).IsInstanceOfType(obj))
			{
				return new DERObjectIdentifier(((ASN1OctetString)obj).getOctets());
			}

			if (typeof(ASN1TaggedObject).IsInstanceOfType(obj))
			{
				return getInstance(((ASN1TaggedObject)obj).getObject());
			}

			throw new ArgumentException("illegal object in getInstance: " + obj.GetType().ToString());
		}

		/**
		* return an Object Identifier from a tagged object.
		*
		* @param obj the tagged object holding the object we want
		* @param explicit true if the object is meant to be explicitly
		*              tagged false otherwise.
		* @exception ArgumentException if the tagged object cannot
		*               be converted.
		*/
		public static DERObjectIdentifier getInstance(
			ASN1TaggedObject obj,
			bool          flagExplicit)
		{
			return getInstance(obj.getObject());
		}
	    

		internal DERObjectIdentifier(byte[]  bytes)
		{
			StringBuilder    objId = new StringBuilder();
			int             value = 0;
			bool         first = true;

			for (int i = 0; i != bytes.Length; i++)
			{
				int b = bytes[i] & 0xff;

				value = value * 128 + (b & 0x7f);
				if ((b & 0x80) == 0)             // end of number reached
				{
					if (first)
					{
						switch (value / 40)
						{
						case 0:
							objId.Append('0');
							break;
						case 1:
							objId.Append('1');
							value -= 40;
							break;
						default:
							objId.Append('2');
							value -= 80;
							break;
						}
						first = false;
					}

					objId.Append('.');
					objId.Append(Convert.ToString(value));
					value = 0;
				}
			}

			this.identifier = objId.ToString();
		}

		public DERObjectIdentifier(
			String  identifier)
		{
			this.identifier = identifier;
		}

		public String getId()
		{
			return identifier;
		}

		private void writeField(
			Stream    outStream,
			int             fieldValue)
			//throws IOException
		{
			if (fieldValue >= (1 << 7))
			{
				if (fieldValue >= (1 << 14))
				{
					if (fieldValue >= (1 << 21))
					{
						if (fieldValue >= (1 << 28))
						{
							outStream.WriteByte((byte)((fieldValue >> 28) | 0x80));
						}
						outStream.WriteByte((byte)((fieldValue >> 21) | 0x80));
					}
					outStream.WriteByte((byte)((fieldValue >> 14) | 0x80));
				}
				outStream.WriteByte((byte)((fieldValue >> 7) | 0x80));
			}
			outStream.WriteByte((byte)(fieldValue & 0x7f));
		}

		public override void encode(DEROutputStream outStream)
		{
			OIDTokenizer            tok = new OIDTokenizer(identifier);
			MemoryStream   bOut = new MemoryStream();
			DEROutputStream         dOut = new DEROutputStream(bOut);

			writeField(bOut, 
						Int32.Parse(tok.nextToken()) * 40
						+ Int32.Parse(tok.nextToken()));

			while (tok.hasMoreTokens())
			{
				writeField(bOut, Int32.Parse(tok.nextToken()));
			}

			dOut.close();

			byte[]  bytes = bOut.ToArray();

			outStream.writeEncoded(DERTags.OBJECT_IDENTIFIER, bytes);
		}

		public override  int GetHashCode()
		{
			return identifier.GetHashCode();
		}

		public override bool Equals(Object  o)
		{
			if ((o == null) || !(typeof(DERObjectIdentifier).IsInstanceOfType(o)))
			{
				return false;
			}

			return identifier.Equals(((DERObjectIdentifier)o).identifier);
		}
	}

}