using System;
using org.bouncycastle.crypto;
using org.bouncycastle.crypto.parameters;

namespace org.bouncycastle.crypto.modes
{
	/**
	* Implements OpenPGP's rather strange version of Cipher-FeedBack (CFB) mode on top of a simple cipher.
	* @see RFC2440
	*/
	public class PGPCFBBlockCipher : BlockCipher
	{
		private byte[] IV;
		private byte[] FR;
		private byte[] FRE;
		private byte[] tmp;

		private BlockCipher cipher;

		private int count;
		private int blockSize;
		private bool forEncryption;
	    
		private bool inlineIv; // if false we don't need to prepend an IV

		/**
		* Basic constructor.
		*
		* @param cipher the block cipher to be used as the basis of the
		* feedback mode.
		* @param inlineIv if true this is for PGP CFB with a prepended iv.
		*/
		public PGPCFBBlockCipher(
			BlockCipher cipher,
			bool     inlineIv)
		{
			this.cipher = cipher;
			this.inlineIv = inlineIv;

			this.blockSize = cipher.getBlockSize();
			if (inlineIv) 
				this.IV = new byte[blockSize + 2];
			else
				this.IV = new byte[blockSize];
			this.FR = new byte[blockSize];
			this.FRE = new byte[blockSize];
			this.tmp = new byte[blockSize];
		}

		/**
		* return the underlying block cipher that we are wrapping.
		*
		* @return the underlying block cipher that we are wrapping.
		*/
		public BlockCipher getUnderlyingCipher()
		{
			return cipher;
		}
	    
		/**
		* return the algorithm name and mode.
		*
		* @return the name of the underlying algorithm followed by "/PGPCFB"
		* and the block size in bits.
		*/
		public String getAlgorithmName()
		{
			if (inlineIv)
			{
				return cipher.getAlgorithmName() + "/PGPCFBwithIV";
			}
			else
			{
				return cipher.getAlgorithmName() + "/PGPCFB";
			}
		}
	    
		/**
		* return the block size we are operating at.
		*
		* @return the block size we are operating at (in bytes).
		*/
		public int getBlockSize()
		{
			return cipher.getBlockSize();
		}

		/**
		* Process one block of input from the array in and write it to
		* the out array.
		*
		* @param in the array containing the input data.
		* @param inOff offset into the in array the data starts at.
		* @param out the array the output data will be copied into.
		* @param outOff the offset into the out array the output will start at.
		* @exception DataLengthException if there isn't enough data in in, or
		* space in out.
		* @exception IllegalStateException if the cipher isn't initialised.
		* @return the number of bytes processed and produced.
		*/
		public int processBlock(
			byte[] inBytes,
			int inOff,
			byte[] outBytes,
			int outOff)
			//throws DataLengthException, IllegalStateException
		{
			if (inlineIv)
			{
				return (forEncryption) ? encryptBlockWithIV(inBytes, inOff, outBytes, outOff) 
									   : decryptBlockWithIV(inBytes, inOff, outBytes, outOff);
			}
			else
			{
				return (forEncryption) ? encryptBlock(inBytes, inOff, outBytes, outOff) 
									   : decryptBlock(inBytes, inOff, outBytes, outOff);
			}
		}
	    
		/**
		* reset the chaining vector back to the IV and reset the underlying
		* cipher.
		*/
		public void reset()
		{
			count = 0;

			for (int i = 0; i != FR.Length; i++)
			{
					if (inlineIv)
						FR[i] = 0;
					else
						FR[i] = IV[i]; // if simple mode, key is IV (even if this is zero)
			}

			cipher.reset();
		}

		/**
		* Initialise the cipher and, possibly, the initialisation vector (IV).
		* If an IV isn't passed as part of the parameter, the IV will be all zeros.
		* An IV which is too short is handled in FIPS compliant fashion.
		*
		* @param forEncryption if true the cipher is initialised for
		*  encryption, if false for decryption.
		* @param param the key and other data required by the cipher.
		* @exception IllegalArgumentException if the params argument is
		* inappropriate.
		*/
		public void init(
			bool forEncryption,
			CipherParameters parameters)
			//throws IllegalArgumentException
		{
			this.forEncryption = forEncryption;
	     
			if (typeof(ParametersWithIV).IsInstanceOfType(parameters))
			{
					ParametersWithIV ivParam = (ParametersWithIV)parameters;
					byte[]      iv = ivParam.getIV();

					if (iv.Length < IV.Length)
					{
						// prepend the supplied IV with zeros (per FIPS PUB 81)
						Array.Copy(iv, 0, IV, IV.Length - iv.Length, iv.Length);
						for (int i = 0; i < IV.Length - iv.Length; i++)
						{
								IV[i] = 0;
						}
					}
					else
					{
						Array.Copy(iv, 0, IV, 0, IV.Length);
					}

					reset();

					cipher.init(true, ivParam.getParameters());
			}
			else
			{
					reset();

					cipher.init(true, parameters);
			}
		}
	    
		/**
		* Encrypt one byte of data according to CFB mode.
		* @param data the byte to encrypt
		* @param where am i in the current block, determines when to resync the block
		* @returns the encrypted byte
		*/
		private byte encryptByte(byte data, int blockOff)
		{
			return (byte)(FRE[blockOff] ^ data);
		}
	    
		/**
		* Do the appropriate processing for CFB IV mode encryption.
		*
		* @param in the array containing the data to be encrypted.
		* @param inOff offset into the in array the data starts at.
		* @param out the array the encrypted data will be copied into.
		* @param outOff the offset into the out array the output will start at.
		* @exception DataLengthException if there isn't enough data in in, or
		* space in out.
		* @exception IllegalStateException if the cipher isn't initialised.
		* @return the number of bytes processed and produced.
		*/
		private int encryptBlockWithIV(
			byte[] inBytes,
			int inOff,
			byte[] outBytes,
			int outOff)
			//throws DataLengthException, IllegalStateException
		{
			if ((inOff + blockSize) > inBytes.Length)
			{
				throw new DataLengthException("input buffer too short");
			}

			if ((outOff + blockSize) > outBytes.Length)
			{
				throw new DataLengthException("output buffer too short");
			}
	        
			if (count == 0)
			{
				cipher.processBlock(FR, 0, FRE, 0);

				for (int n = 0; n < blockSize; n++) 
				{
					outBytes[outOff + n] = encryptByte(IV[n], n);
				}
	            
				Array.Copy(outBytes, outOff, FR, 0, blockSize);

				cipher.processBlock(FR, 0, FRE, 0);

				outBytes[outOff + blockSize] = encryptByte(IV[blockSize], 0);
				outBytes[outOff + blockSize + 1] = encryptByte(IV[blockSize + 1], 1);

				Array.Copy(outBytes, outOff + 2, FR, 0, blockSize);
	            
				cipher.processBlock(FR, 0, FRE, 0);

				for (int n = 0; n < blockSize; n++) 
				{
					outBytes[outOff + blockSize + 2 + n] = encryptByte(inBytes[inOff + n], n);
				}

				Array.Copy(outBytes, outOff + blockSize + 2, FR, 0, blockSize);

				count += 2 * blockSize + 2;

				return 2 * blockSize + 2;
			}
			else if (count >= blockSize + 2)
			{
				cipher.processBlock(FR, 0, FRE, 0);

				for (int n = 0; n < blockSize; n++) 
				{
					outBytes[outOff + n] = encryptByte(inBytes[inOff + n], n);
				}
	            
				Array.Copy(outBytes, outOff, FR, 0, blockSize);
			}
	        
			return blockSize;
		}

		/**
		* Do the appropriate processing for CFB IV mode decryption.
		*
		* @param in the array containing the data to be decrypted.
		* @param inOff offset into the in array the data starts at.
		* @param out the array the encrypted data will be copied into.
		* @param outOff the offset into the out array the output will start at.
		* @exception DataLengthException if there isn't enough data in in, or
		* space in out.
		* @exception IllegalStateException if the cipher isn't initialised.
		* @return the number of bytes processed and produced.
		*/
		private int decryptBlockWithIV(
			byte[] inBytes,
			int inOff,
			byte[] outBytes,
			int outOff)
			//throws DataLengthException, IllegalStateException
		{
			if ((inOff + blockSize) > inBytes.Length)
			{
				throw new DataLengthException("input buffer too short");
			}

			if ((outOff + blockSize) > outBytes.Length)
			{
				throw new DataLengthException("output buffer too short");
			}
	        
			if (count == 0)
			{
				for (int n = 0; n < blockSize; n++) 
				{
					FR[n] = inBytes[inOff + n];
				}
	            
				cipher.processBlock(FR, 0, FRE, 0);

				count += blockSize;

				return 0;
			}
			else if (count == blockSize)
			{
				// copy in buffer so that this mode works if in and out are the same 
				Array.Copy(inBytes, inOff, tmp, 0, blockSize);
	        
				Array.Copy(FR, 2, FR, 0, blockSize - 2);
	            
				FR[blockSize - 2] = tmp[0];
				FR[blockSize - 1] = tmp[1];

				cipher.processBlock(FR, 0, FRE, 0);

				for (int n = 0; n < blockSize - 2; n++) 
				{
					outBytes[outOff + n] = encryptByte(tmp[n + 2], n);
				}

				Array.Copy(tmp, 2, FR, 0, blockSize - 2);

				count += 2;

				return blockSize - 2;
			}
			else if (count >= blockSize + 2)
			{
				// copy in buffer so that this mode works if in and out are the same 
				Array.Copy(inBytes, inOff, tmp, 0, blockSize);

				outBytes[outOff + 0] = encryptByte(tmp[0], blockSize - 2);
				outBytes[outOff + 1] = encryptByte(tmp[1], blockSize - 1);

				Array.Copy(tmp, 0, FR, blockSize - 2, 2);

				cipher.processBlock(FR, 0, FRE, 0);
	            
				for (int n = 0; n < blockSize - 2; n++) 
				{
					outBytes[outOff + n + 2] = encryptByte(tmp[n + 2], n);
				}
	            
				Array.Copy(tmp, 2, FR, 0, blockSize - 2);
	            
			} 
	        
			return blockSize;
		}
	    
		/**
		* Do the appropriate processing for CFB mode encryption.
		*
		* @param in the array containing the data to be encrypted.
		* @param inOff offset into the in array the data starts at.
		* @param out the array the encrypted data will be copied into.
		* @param outOff the offset into the out array the output will start at.
		* @exception DataLengthException if there isn't enough data in in, or
		* space in out.
		* @exception IllegalStateException if the cipher isn't initialised.
		* @return the number of bytes processed and produced.
		*/
		private int encryptBlock(
			byte[] inBytes,
			int inOff,
			byte[] outBytes,
			int outOff)
			//throws DataLengthException, IllegalStateException
		{        
			if ((inOff + blockSize) > inBytes.Length)
			{
				throw new DataLengthException("input buffer too short");
			}

			if ((outOff + blockSize) > outBytes.Length)
			{
				throw new DataLengthException("output buffer too short");
			}
	        
	        
			cipher.processBlock(FR, 0, FRE, 0);
			for (int n = 0; n < blockSize; n++) 
				outBytes[outOff + n] = encryptByte(inBytes[inOff + n], n);
	        
			for (int n = 0; n < blockSize; n++) 
				FR[n] = outBytes[outOff + n];
	        
			return blockSize;
	        
		}
	    
		/**
		* Do the appropriate processing for CFB mode decryption.
		*
		* @param in the array containing the data to be decrypted.
		* @param inOff offset into the in array the data starts at.
		* @param out the array the encrypted data will be copied into.
		* @param outOff the offset into the out array the output will start at.
		* @exception DataLengthException if there isn't enough data in in, or
		* space in out.
		* @exception IllegalStateException if the cipher isn't initialised.
		* @return the number of bytes processed and produced.
		*/
		private int decryptBlock(
			byte[] inBytes,
			int inOff,
			byte[] outBytes,
			int outOff)
			//throws DataLengthException, IllegalStateException
		{
			if ((inOff + blockSize) > inBytes.Length)
			{
				throw new DataLengthException("input buffer too short");
			}

			if ((outOff + blockSize) > outBytes.Length)
			{
				throw new DataLengthException("output buffer too short");
			}
	        
			cipher.processBlock(FR, 0, FRE, 0);
			for (int n = 0; n < blockSize; n++) 
				outBytes[outOff + n] = encryptByte(inBytes[inOff + n], n);
	        
			for (int n = 0; n < blockSize; n++) 
				FR[n] = inBytes[inOff + n];
	        
			return blockSize;
	        
		}
	}

}